@@ -36,6 +36,30 @@ static int use_lt_credentials = 0;
 static int use_st_credentials = 0;
 static int anon_credentials = 0;
 
+////// TURNDB //////////////
+
+#if defined(TURNDB)
+
+#if defined(Q)
+#undef Q
+#endif
+
+#define Q(x) #x
+
+#if defined(QUOTE)
+#undef QUOTE
+#endif
+
+#define QUOTE(x) Q(x)
+
+#define DEFAULT_USERDB_FILE QUOTE(TURNDB)
+
+#else
+
+#define DEFAULT_USERDB_FILE "/usr/local/var/db/turndb"
+
+#endif
+
 //////TURN PARAMS STRUCTURE DEFINITION //////
 
 #define DEFAULT_GENERAL_RELAY_SERVERS_NUMBER (1)
@@ -77,7 +101,7 @@ DEFAULT_STUN_PORT,DEFAULT_STUN_TLS_PORT,0,0,1,
 {
   NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,0,0,NULL,NULL,NULL
 },
-{NULL, NULL, 0},{NULL, NULL, 0},
+{NULL, 0},{NULL, 0},
 NEV_UNKNOWN, 
 { "Unknown", "UDP listening socket per session", "UDP thread per network endpoint", "UDP thread per CPU core" },
 //////////////// Relay servers //////////////////////////////////
@@ -95,7 +119,7 @@ LOW_DEFAULT_PORTS_BOUNDARY,HIGH_DEFAULT_PORTS_BOUNDARY,0,0,0,"",
 /////////////// MISC PARAMS ////////////////
 0,0,0,0,0,SHATYPE_SHA1,':',0,0,TURN_CREDENTIALS_NONE,0,0,0,0,0,0,
 ///////////// Users DB //////////////
-{ TURN_USERDB_TYPE_FILE, {"\0",NULL}, {0,NULL,NULL, {NULL,0}} }
+{ (TURN_USERDB_TYPE)0, {"\0",NULL}, {0,NULL,NULL, {NULL,0}} }
 
 };
 
@@ -374,7 +398,8 @@ static char Usage[] = "Usage: turnserver [options]\n"
 " --no-multicast-peers				Disallow peers on well-known broadcast addresses (224.0.0.0 and above, and FFXX:*).\n"
 " -m, --relay-threads		<number>	Number of relay threads to handle the established connections\n"
 "						(in addition to authentication thread and the listener thread).\n"
-"						If set to 0 then application runs in single-threaded mode.\n"
+"						If explicitly set to 0 then application runs in single-threaded mode.\n"
+"						If not set then a default OS-dependent optimal algorithm will be employed.\n"
 "						The default thread number is the number of CPUs.\n"
 "						In older systems (pre-Linux 3.9) the number of UDP relay threads always equals\n"
 "						the number of listening endpoints (unless -m 0 is set).\n"
@@ -386,17 +411,14 @@ static char Usage[] = "Usage: turnserver [options]\n"
 " -V, --Verbose					Extra verbose mode, very annoying (for debug purposes only).\n"
 " -o, --daemon					Start process as daemon (detach from current shell).\n"
 " -f, --fingerprint				Use fingerprints in the TURN messages.\n"
-" -a, --lt-cred-mech				Use the long-term credential mechanism. This option can be used with either\n"
-"		                                flat file user database or PostgreSQL DB or MySQL DB for user keys storage.\n"
-" -A, --st-cred-mech				Use the short-term credential mechanism. This option requires\n"
-"		                                a PostgreSQL or MySQL DB for short term passwords storage.\n"
+" -a, --lt-cred-mech				Use the long-term credential mechanism.\n"
+" -A, --st-cred-mech				Use the short-term credential mechanism.\n"
 " -z, --no-auth					Do not use any credential mechanism, allow anonymous access.\n"
 " -u, --user			<user:pwd>	User account, in form 'username:password', for long-term credentials.\n"
 "						Cannot be used with TURN REST API or with short-term credentials.\n"
 " -r, --realm			<realm>		The default realm to be used for the users when no explicit\n"
-"						origin/realm relationship was found in the database, or if the TURN\n"
-"						server is not using any database (just the commands-line settings\n"
-"						and the userdb file). Must be used with long-term credentials \n"
+"						origin/realm relationship was found in the database.\n"
+"						Must be used with long-term credentials \n"
 "						mechanism or with TURN REST API.\n"
 " --check-origin-consistency			The flag that sets the origin consistency check:\n"
 "						across the session, all requests must have the same\n"
@@ -415,7 +437,10 @@ static char Usage[] = "Usage: turnserver [options]\n"
 "						Total bytes-per-second bandwidth the TURN server is allowed to allocate\n"
 "						for the sessions, combined (input and output network streams are treated separately).\n"
 " -c				<filename>	Configuration file name (default - turnserver.conf).\n"
-" -b, --userdb			<filename>	User database file name (default - turnuserdb.conf) for long-term credentials only.\n"
+#if !defined(TURN_NO_SQLITE)
+" -b, , --db, --userdb	<filename>		SQLite database file name; default - /var/db/turndb or\n"
+"						    /usr/local/var/db/turndb or /var/lib/turn/turndb.\n"
+#endif
 #if !defined(TURN_NO_PQ)
 " -e, --psql-userdb, --sql-userdb <conn-string>	PostgreSQL database connection string, if used (default - empty, no PostreSQL DB used).\n"
 "		                                This database can be used for long-term and short-term credentials mechanisms,\n"
@@ -569,7 +594,7 @@ static char Usage[] = "Usage: turnserver [options]\n"
 "\n";
 
 static char AdminUsage[] = "Usage: turnadmin [command] [options]\n"
-	"Commands:\n"
+	"\nCommands:\n\n"
 	"	-k, --key			generate long-term credential mechanism key for a user\n"
 	"	-a, --add			add/update a long-term mechanism user\n"
 	"	-A, --add-st			add/update a short-term mechanism user\n"
@@ -577,7 +602,6 @@ static char AdminUsage[] = "Usage: turnadmin [command] [options]\n"
 	"	-D, --delete-st			delete a short-term mechanism user\n"
 	"	-l, --list			list all long-term mechanism users\n"
 	"	-L, --list-st			list all short-term mechanism users\n"
-#if !defined(TURN_NO_PQ) || !defined(TURN_NO_MYSQL) || !defined(TURN_NO_MONGO) || !defined(TURN_NO_HIREDIS)
 	"	-s, --set-secret=<value>	Add shared secret for TURN RESP API\n"
 	"	-S, --show-secret		Show stored shared secrets for TURN REST API\n"
 	"	-X, --delete-secret=<value>	Delete a shared secret\n"
@@ -587,9 +611,11 @@ static char AdminUsage[] = "Usage: turnadmin [command] [options]\n"
 	"	-I, --list-origins		List origin-to-realm relations.\n"
 	"	-g, --set-realm-option		Set realm params: max-bps, total-quota, user-quota.\n"
 	"	-G, --list-realm-options	List realm params.\n"
+	"\nOptions with mandatory values:\n\n"
+#if !defined(TURN_NO_SQLITE)
+	"	-b, --db, --userdb		SQLite database file, default value is /var/db/turndb or\n"
+	"					  /usr/local/var/db/turndb or /var/lib/turn/turndb.\n"
 #endif
-	"Options with mandatory values:\n"
-	"	-b, --userdb			User database file, if flat DB file is used.\n"
 #if !defined(TURN_NO_PQ)
 	"	-e, --psql-userdb, --sql-userdb	PostgreSQL user database connection string, if PostgreSQL DB is used.\n"
 #endif
@@ -605,7 +631,7 @@ static char AdminUsage[] = "Usage: turnadmin [command] [options]\n"
 	"	-u, --user			Username\n"
 	"	-r, --realm			Realm for long-term mechanism only\n"
 	"	-p, --password			Password\n"
-#if !defined(TURN_NO_PQ) || !defined(TURN_NO_MYSQL) || !defined(TURN_NO_MONGO) || !defined(TURN_NO_HIREDIS)
+#if !defined(TURN_NO_SQLITE) || !defined(TURN_NO_PQ) || !defined(TURN_NO_MYSQL) || !defined(TURN_NO_MONGO) || !defined(TURN_NO_HIREDIS)
 	"	-o, --origin			Origin\n"
 #endif
 	"	-H, --sha256			Use SHA256 digest function to be used for the message integrity.\n"
@@ -719,7 +745,10 @@ static const struct myoption long_options[] = {
 				{ "st-cred-mech", optional_argument, NULL, 'A' },
 				{ "no-auth", optional_argument, NULL, 'z' },
 				{ "user", required_argument, NULL, 'u' },
+#if !defined(TURN_NO_SQLITE)
 				{ "userdb", required_argument, NULL, 'b' },
+				{ "db", required_argument, NULL, 'b' },
+#endif
 #if !defined(TURN_NO_PQ)
 				{ "psql-userdb", required_argument, NULL, 'e' },
 				{ "sql-userdb", required_argument, NULL, 'e' },
@@ -808,15 +837,16 @@ static const struct myoption admin_long_options[] = {
 				{ "delete", no_argument, NULL, 'd' },
 				{ "list", no_argument, NULL, 'l' },
 				{ "list-st", no_argument, NULL, 'L' },
-#if !defined(TURN_NO_PQ) || !defined(TURN_NO_MYSQL) || !defined(TURN_NO_MONGO) || !defined(TURN_NO_HIREDIS)
 				{ "set-secret", required_argument, NULL, 's' },
 				{ "show-secret", no_argument, NULL, 'S' },
 				{ "delete-secret", required_argument, NULL, 'X' },
 				{ "delete-all-secrets", no_argument, NULL, DEL_ALL_AUTH_SECRETS_OPT },
-#endif
 				{ "add-st", no_argument, NULL, 'A' },
 				{ "delete-st", no_argument, NULL, 'D' },
+#if !defined(TURN_NO_SQLITE)
 				{ "userdb", required_argument, NULL, 'b' },
+				{ "db", required_argument, NULL, 'b' },
+#endif
 #if !defined(TURN_NO_PQ)
 				{ "psql-userdb", required_argument, NULL, 'e' },
 				{ "sql-userdb", required_argument, NULL, 'e' },
@@ -834,7 +864,6 @@ static const struct myoption admin_long_options[] = {
 				{ "realm", required_argument, NULL, 'r' },
 				{ "password", required_argument, NULL, 'p' },
 				{ "sha256", no_argument, NULL, 'H' },
-#if !defined(TURN_NO_PQ) || !defined(TURN_NO_MYSQL) || !defined(TURN_NO_MONGO) || !defined(TURN_NO_HIREDIS)
 				{ "add-origin", no_argument, NULL, 'O' },
 				{ "del-origin", no_argument, NULL, 'R' },
 				{ "list-origins", required_argument, NULL, 'I' },
@@ -844,7 +873,6 @@ static const struct myoption admin_long_options[] = {
 				{ "user-quota", required_argument, NULL, ADMIN_USER_QUOTA_OPT },
 				{ "total-quota", required_argument, NULL, ADMIN_TOTAL_QUOTA_OPT },
 				{ "max-bps", required_argument, NULL, ADMIN_MAX_BPS_OPT },
-#endif
 				{ "help", no_argument, NULL, 'h' },
 				{ NULL, no_argument, NULL, 0 }
 };
@@ -1111,10 +1139,12 @@ static void set_option(int c, char *value)
 	case 'u':
 		add_user_account(value,0);
 		break;
+#if !defined(TURN_NO_SQLITE)
 	case 'b':
 		STRCPY(turn_params.default_users_db.persistent_users_db.userdb, value);
-		turn_params.default_users_db.userdb_type = TURN_USERDB_TYPE_FILE;
+		turn_params.default_users_db.userdb_type = TURN_USERDB_TYPE_SQLITE;
 		break;
+#endif
 #if !defined(TURN_NO_PQ)
 	case 'e':
 		STRCPY(turn_params.default_users_db.persistent_users_db.userdb, value);
@@ -1230,10 +1260,10 @@ static void set_option(int c, char *value)
 		add_tls_alternate_server(value);
 		break;
 	case ALLOWED_PEER_IPS:
-		if (add_ip_list_range(value, &turn_params.ip_whitelist) == 0) TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "White listing: %s\n", value);
+		if (add_ip_list_range(value, NULL, &turn_params.ip_whitelist) == 0) TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "White listing: %s\n", value);
 		break;
 	case DENIED_PEER_IPS:
-		if (add_ip_list_range(value, &turn_params.ip_blacklist) == 0) TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "Black listing: %s\n", value);
+		if (add_ip_list_range(value, NULL, &turn_params.ip_blacklist) == 0) TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "Black listing: %s\n", value);
 		break;
 	case CIPHER_LIST_OPT:
 		STRCPY(turn_params.cipher_list,value);
@@ -1461,7 +1491,6 @@ static int adminmain(int argc, char **argv)
 			ct = TA_LIST_USERS;
 			is_st = 1;
 			break;
-#if !defined(TURN_NO_PQ) || !defined(TURN_NO_MYSQL) || !defined(TURN_NO_MONGO) || !defined(TURN_NO_HIREDIS)
 		case 's':
 			ct = TA_SET_SECRET;
 			STRCPY(secret,optarg);
@@ -1477,11 +1506,12 @@ static int adminmain(int argc, char **argv)
 		case DEL_ALL_AUTH_SECRETS_OPT:
 			ct = TA_DEL_SECRET;
 			break;
-#endif
+#if !defined(TURN_NO_SQLITE)
 		case 'b':
 		  STRCPY(turn_params.default_users_db.persistent_users_db.userdb,optarg);
-		  turn_params.default_users_db.userdb_type = TURN_USERDB_TYPE_FILE;
+		  turn_params.default_users_db.userdb_type = TURN_USERDB_TYPE_SQLITE;
 		  break;
+#endif
 #if !defined(TURN_NO_PQ)
 		case 'e':
 		  STRCPY(turn_params.default_users_db.persistent_users_db.userdb,optarg);
@@ -1544,13 +1574,10 @@ static int adminmain(int argc, char **argv)
 		}
 	}
 
-	if(is_st && (turn_params.default_users_db.userdb_type == TURN_USERDB_TYPE_FILE)) {
-		TURN_LOG_FUNC(TURN_LOG_LEVEL_ERROR, "ERROR: you have to use a PostgreSQL or MySQL database with short-term credentials\n");
-		exit(-1);
-	}
-
-	if(!strlen(turn_params.default_users_db.persistent_users_db.userdb) && (turn_params.default_users_db.userdb_type == TURN_USERDB_TYPE_FILE))
+#if !defined(TURN_NO_SQLITE)
+	if(!strlen(turn_params.default_users_db.persistent_users_db.userdb) && (turn_params.default_users_db.userdb_type == TURN_USERDB_TYPE_SQLITE))
 		STRCPY(turn_params.default_users_db.persistent_users_db.userdb,DEFAULT_USERDB_FILE);
+#endif
 
 	if(ct == TA_COMMAND_UNKNOWN) {
 		fprintf(stderr,"\n%s\n", AdminUsage);
@@ -1601,6 +1628,12 @@ static void print_features(unsigned long mfn)
 	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "AEAD supported\n");
 #endif
 
+#if !defined(TURN_NO_SQLITE)
+	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "SQLite supported, default database location is %s\n",DEFAULT_USERDB_FILE);
+#else
+	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "SQLite is not supported\n");
+#endif
+
 #if !defined(TURN_NO_HIREDIS)
 	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "Redis supported\n");
 #else
@@ -1701,9 +1734,11 @@ static void drop_privileges(void)
 static void init_domain(void)
 {
 #if !defined(TURN_NO_GETDOMAINNAME)
-	getdomainname(turn_params.domain,sizeof(turn_params.domain)-1);
-	if(!strcmp(turn_params.domain,"(none)")) 
-	  turn_params.domain[0]=0;
+	if(getdomainname(turn_params.domain,sizeof(turn_params.domain)-1)<0) {
+		turn_params.domain[0]=0;
+	} else if(!strcmp(turn_params.domain,"(none)")) {
+		turn_params.domain[0]=0;
+	}
 #endif
 }
 
@@ -1849,10 +1884,11 @@ int main(int argc, char **argv)
 		TURN_LOG_FUNC(TURN_LOG_LEVEL_WARNING, "\nCONFIG: WARNING: --server-relay: NON-STANDARD AND DANGEROUS OPTION.\n");
 	}
 
-	if(!strlen(turn_params.default_users_db.persistent_users_db.userdb) && (turn_params.default_users_db.userdb_type == TURN_USERDB_TYPE_FILE))
+#if !defined(TURN_NO_SQLITE)
+	if(!strlen(turn_params.default_users_db.persistent_users_db.userdb) && (turn_params.default_users_db.userdb_type == TURN_USERDB_TYPE_SQLITE))
 			STRCPY(turn_params.default_users_db.persistent_users_db.userdb,DEFAULT_USERDB_FILE);
+#endif
 
-	read_userdb_file(0);
 	update_white_and_black_lists();
 
 	argc -= optind;
@@ -1889,9 +1925,7 @@ int main(int argc, char **argv)
 	}
 
 	if(use_lt_credentials) {
-		if(!turn_params.default_users_db.ram_db.users_number && (turn_params.default_users_db.userdb_type == TURN_USERDB_TYPE_FILE) && !turn_params.use_auth_secret_with_timestamp) {
-			TURN_LOG_FUNC(TURN_LOG_LEVEL_WARNING, "\nCONFIGURATION ALERT: you did not specify any user account, (-u option) \n	but you did specified a long-term credentials mechanism option (-a option).\n	The TURN Server will be inaccessible.\n		Check your configuration.\n");
-		} else if(!get_realm(NULL)->options.name[0]) {
+		if(!get_realm(NULL)->options.name[0]) {
 			TURN_LOG_FUNC(TURN_LOG_LEVEL_WARNING, "\nCONFIGURATION ALERT: you did specify the long-term credentials usage\n but you did not specify the default realm option (-r option).\n		Check your configuration.\n");
 		}
 	}
@@ -1960,6 +1994,23 @@ int main(int argc, char **argv)
 		}
 	}
 
+	if(turn_params.external_ip && turn_params.relay_addrs) {
+		size_t ir = 0;
+		for(ir = 0; ir < turn_params.relays_number; ++ir) {
+			if(turn_params.relay_addrs[ir]) {
+				const char* sra = (const char*)turn_params.relay_addrs[ir];
+				if((strstr(sra,"127.0.0.1") != sra)&&(strstr(sra,"::1")!=sra)) {
+					ioa_addr ra;
+					if(make_ioa_addr((const u08bits*)sra,0,&ra)<0) {
+						TURN_LOG_FUNC(TURN_LOG_LEVEL_ERROR,"-X : Wrong address format: %s\n",sra);
+					} else if(ra.ss.sa_family == turn_params.external_ip->ss.sa_family) {
+						ioa_addr_add_mapping(turn_params.external_ip,&ra);
+					}
+				}
+			}
+		}
+	}
+
 	if(turn_params.turn_daemon) {
 #if !defined(TURN_HAS_DAEMON)
 		pid_t pid = fork();
@@ -2030,15 +2081,21 @@ int main(int argc, char **argv)
 
 #if defined(OPENSSL_THREADS)
 
-static pthread_mutex_t* mutex_buf = NULL;
+static char some_buffer[65536];
+
+//array larger than anything that OpenSSL may need:
+static pthread_mutex_t mutex_buf[256];
+static int mutex_buf_initialized = 0;
 
 static void locking_function(int mode, int n, const char *file, int line) {
   UNUSED_ARG(file);
   UNUSED_ARG(line);
-  if (mode & CRYPTO_LOCK)
-    pthread_mutex_lock(&mutex_buf[n]);
-  else
-    pthread_mutex_unlock(&mutex_buf[n]);
+  if(mutex_buf_initialized && (n < CRYPTO_num_locks())) {
+	  if (mode & CRYPTO_LOCK)
+		  pthread_mutex_lock(&(mutex_buf[n]));
+	  else
+		  pthread_mutex_unlock(&(mutex_buf[n]));
+  }
 }
 
 #if OPENSSL_VERSION_NUMBER >= 0x10000000L
@@ -2061,12 +2118,13 @@ static int THREAD_setup(void) {
 
 	int i;
 
-	mutex_buf = (pthread_mutex_t*) turn_malloc(CRYPTO_num_locks()
-			* sizeof(pthread_mutex_t));
-	if (!mutex_buf)
-		return 0;
-	for (i = 0; i < CRYPTO_num_locks(); i++)
-		pthread_mutex_init(&mutex_buf[i], NULL);
+	some_buffer[0] = 0;
+
+	for (i = 0; i < CRYPTO_num_locks(); i++) {
+		pthread_mutex_init(&(mutex_buf[i]), NULL);
+	}
+
+	mutex_buf_initialized = 1;
 
 #if OPENSSL_VERSION_NUMBER >= 0x10000000L
 	CRYPTO_THREADID_set_callback(id_function);
@@ -2087,7 +2145,7 @@ int THREAD_cleanup(void) {
 
   int i;
 
-  if (!mutex_buf)
+  if (!mutex_buf_initialized)
     return 0;
 
 #if OPENSSL_VERSION_NUMBER >= 0x10000000L
@@ -2097,10 +2155,11 @@ int THREAD_cleanup(void) {
 #endif
 
   CRYPTO_set_locking_callback(NULL);
-  for (i = 0; i < CRYPTO_num_locks(); i++)
-    pthread_mutex_destroy(&mutex_buf[i]);
-  turn_free(mutex_buf,sizeof(pthread_mutex_t));
-  mutex_buf = NULL;
+  for (i = 0; i < CRYPTO_num_locks(); i++) {
+	  pthread_mutex_destroy(&(mutex_buf[i]));
+  }
+
+  mutex_buf_initialized = 0;
 
 #endif
 
