@@ -48,8 +48,6 @@ extern "C" {
 
 //////////// Defines //////////////////////////////
 
-#define DEFAULT_USERDB_FILE "turnuserdb.conf"
-
 #define AUTH_SECRET_SIZE (512)
 
 //////////// REALM //////////////
@@ -100,7 +98,12 @@ struct auth_message {
 };
 
 enum _TURN_USERDB_TYPE {
-	TURN_USERDB_TYPE_FILE=0
+#if !defined(TURN_NO_SQLITE)
+	TURN_USERDB_TYPE_UNKNOWN=-1,
+	TURN_USERDB_TYPE_SQLITE=0
+#else
+	TURN_USERDB_TYPE_UNKNOWN=0
+#endif
 #if !defined(TURN_NO_PQ)
 	,TURN_USERDB_TYPE_PQ
 #endif
@@ -202,13 +205,14 @@ void release_allocation_quota(u08bits *username, int oauth, u08bits *realm);
 	void run_db_test(void);
 #endif
 
-void read_userdb_file(int to_print);
 void auth_ping(redis_context_handle rch);
 void reread_realms(void);
 int add_user_account(char *user, int dynamic);
 int adminuser(u08bits *user, u08bits *realm, u08bits *pwd, u08bits *secret, u08bits *origin, TURNADMIN_COMMAND_TYPE ct, int is_st, perf_options_t* po);
 
-int add_ip_list_range(const char* range, ip_range_list_t * list);
+int add_ip_list_range(const char* range, const char* realm, ip_range_list_t * list);
+ip_range_list_t* get_ip_list(const char *kind);
+void ip_list_free(ip_range_list_t *l);
 
 ///////////// Redis //////////////////////
 
