@@ -137,16 +137,15 @@ static char Usage[] =
   "	-I	Do not set permissions on TURN relay endpoints\n"
   "		(for testing the non-standard server relay functionality).\n"
   "	-G	Generate extra requests (create permissions, channel bind).\n"
-  " -B  Random disconnect after a few initial packets.\n"
-  " -Z  Dual allocation.\n"
-  " -J	Use oAuth with default test key kid='north'.\n"
+  "	-B	Random disconnect after a few initial packets.\n"
+  "	-Z	Dual allocation.\n"
+  "	-J	Use oAuth with default test key kid='north'.\n"
   "Options:\n"
   "	-l	Message length (Default: 100 Bytes).\n"
   "	-i	Certificate file (for secure connections only, optional).\n"
   "	-k	Private key file (for secure connections only).\n"
   "	-E	CA file for server certificate verification, \n"
   "		if the server certificate to be verified.\n"
-  "	-F	Cipher suite for TLS/DTLS. Default value is DEFAULT.\n"
   "	-p	TURN server port (Default: 3478 unsecure, 5349 secure).\n"
   "	-n	Number of messages to send (Default: 5).\n"
   "	-d	Local interface device (optional).\n"
@@ -188,7 +187,7 @@ void recalculate_restapi_hmac(void) {
 					g_upwd[pwd_length] = 0;
 				}
 			}
-			free(pwd);
+			turn_free(pwd,strlen(pwd)+1);
 		}
 	}
 }
@@ -365,6 +364,7 @@ int main(int argc, char **argv)
 		case 'P':
 			passive_tcp = 1;
 			/* implies 'T': */
+			/* no break */
 		case 'T':
 			relay_transport = STUN_ATTRIBUTE_TRANSPORT_TCP_VALUE;
 			break;
